# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import re

from .event import EventHook
from .messages.error import ErrorMessage
from .tornado.response import StatusCode

WILDCHAR_REXP = re.compile(r'[{}/\\:\s.]+')


class StatsdHook(EventHook):
    """Collect Statsd information in the tchannel req/resp."""

    def __init__(self, statsd):
        """

        :param statsd: instance of `StatsD <https://github.com/etsy/statsd>`
        """
        self._statsd = statsd

    def before_send_request(self, request):
        statsd_name = "tchannel.outbound.calls.sent"
        key = common_prefix(statsd_name, request)

        self._statsd.count(key, 1)

    def after_receive_response(self, request, response):
        if response.code == StatusCode.ok:
            statsd_name = "tchannel.outbound.calls.success"
        else:
            statsd_name = "tchannel.outbound.calls.app-errors"
        key = common_prefix(statsd_name, request)

        self._statsd.count(key, 1)

    def after_receive_system_error(self, request, error):
        statsd_name = "tchannel.outbound.calls.system-errors"
        prefix = common_prefix(statsd_name, request)
        key = prefix + '.' + clean(
            ErrorMessage.ERROR_CODES.get(error.code, None), 'type'
        )

        self._statsd.count(key, 1)

    def after_receive_system_error_per_attempt(self, request, error):
        statsd_name = "tchannel.outbound.calls.per-attempt.system-errors"
        prefix = common_prefix(statsd_name, request)
        key = prefix + '.' + clean(
            ErrorMessage.ERROR_CODES.get(error.code, None), 'type'
        )

        self._statsd.count(key, 1)

    def on_operational_error_per_attempt(self, request, error):
        statsd_name = "tchannel.outbound.calls.per-attempt.operational-errors"
        prefix = common_prefix(statsd_name, request)
        key = prefix + '.' + clean(
            ErrorMessage.ERROR_CODES.get(error.code, None), 'type'
        )

        self._statsd.count(key, 1)

    def on_operational_error(self, request, error):
        statsd_name = "tchannel.outbound.calls.operational-errors"

        prefix = common_prefix(statsd_name, request)
        key = prefix + '.' + clean(
            ErrorMessage.ERROR_CODES.get(error.code, None), 'type'
        )

        self._statsd.count(key, 1)


def extract_metadata(request):
    service = request.headers.get('cn', None)
    target_service = request.service
    target_endpoint = request.endpoint

    return (service, target_service, target_endpoint)


def common_prefix(statsd_name, request):
    (service, target_service, target_endpoint) = (
        extract_metadata(request)
    )

    return '.'.join([statsd_name,
                     clean(service, 'service'),
                     clean(target_service, 'target-service'),
                     clean(target_endpoint, 'target-endpoint')
                     ])


def clean(key, field):
    if not key:
        return 'no-' + field
    else:

        return WILDCHAR_REXP.sub('-', key)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import sys


class RankCalculator(object):
    """RankCalculator calculates the rank of a peer."""

    def get_rank(self, peer):
        raise NotImplementedError()


class PreferIncomingCalculator(RankCalculator):

    # TIERS lists three ranges for three different kinds of peers.
    # 0: ephemeral peers or unconnected peers
    # 1: peers with only outgoing connections
    # 2: peers with incoming connections
    TIERS = [sys.maxint, sys.maxint / 2, 0]

    def get_rank(self, peer):
        """Calculate the peer rank based on connections.

        If the peer has no incoming connections, it will have largest rank.
        In our peer selection strategy, the largest number has least priority
        in the heap.

        If the peer has incoming connections, we will return number of outbound
        pending requests and responses.

        :param peer: instance of `tchannel.tornado.peer.Peer`
        :return: rank of the peer
        """
        if peer.is_ephemeral or not peer.connections:
            return self.TIERS[0]

        if not peer.has_incoming_connections:
            return self.TIERS[1] + peer.total_outbound_pendings

        return self.TIERS[2] + peer.total_outbound_pendings



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from . import common
from .. import rw
from ..glossary import DEFAULT_TIMEOUT
from .base import BaseMessage


class CancelMessage(BaseMessage):
    __slots__ = BaseMessage.__slots__ + (
        'ttl',
        'tracing',
        'why',
    )

    def __init__(self, ttl=DEFAULT_TIMEOUT, tracing=None, why=None, id=0):
        super(CancelMessage, self).__init__(id)
        self.ttl = ttl
        self.tracing = tracing or common.Tracing(0, 0, 0, 0)
        self.why = why or ''


cancel_rw = rw.instance(
    CancelMessage,
    ('ttl', rw.number(4)),                          # ttl:4
    ('tracing', common.tracing_rw),                 # tracing:24
    ('why', rw.len_prefixed_string(rw.number(2))),  # why:2
)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import random

from .container import heap
from .container.heap import HeapOperation
from .container.heap import NoMatchError
from .container.heap import smallest


class PeerHeap(HeapOperation):
    """PeerHeap maintains a min-heap of peers based on their ranks.

    Peer in the heap will be arranged based on the peer's rank and peer's
    order. Order is equal to peer heap's current order number plus random value
    within the current peer size. It solves two problems when peers are in
    same rank:

    Dead peers: If the order is completely random, then an unlucky peer
    with a very bad assigned order may never get selected.

    Determinism: If the insertion order is used as-is, then all TChannel
    instances would follow the same selection pattern, causing load
    imbalance. For example if they get the same static list of peers, they
    will all pick the first one, then the second one, and so on, cycling
    between which host gets overloaded with requests.

    All in all, it will keep certain level randomization but at the same
    time make the peer rank not deterministic among different tchannel
    instances.

    """

    __slots__ = ('peers',)

    def __init__(self):
        self.peers = []
        self.order = 0

    def size(self):
        return len(self.peers)

    def lt(self, i, j):
        """Compare the priority of two peers.

        Primary comparator will be the rank of each peer. If the ``rank`` is
        same then compare the ``order``. The ``order`` attribute of the peer
        tracks the heap push order of the peer. This help solve the imbalance
        problem caused by randomization when deal with same rank situation.

        :param i: ith peer
        :param j: jth peer
        :return: True or False
        """
        if self.peers[i].rank == self.peers[j].rank:
            return self.peers[i].order < self.peers[j].order

        return self.peers[i].rank < self.peers[j].rank

    def peek(self, i):
        return self.peers[i]

    def push(self, x):
        x.index = len(self.peers)
        self.peers.append(x)

    def pop(self):
        item = self.peers.pop()
        return item

    def swap(self, i, j):
        self.peers[i], self.peers[j] = self.peers[j], self.peers[i]
        self.peers[i].index = i
        self.peers[j].index = j

    def update_peer(self, peer):
        """Update the peer's position in the heap after peer's rank changed"""
        heap.fix(self, peer.index)

    def pop_peer(self):
        """Pop the top peer from the heap

        :return
            return the top peer and remove it from the heap if heap is not
            empty.Otherwise return None.
        """
        if not self.peers:
            return None
        return heap.pop(self)

    def push_peer(self, peer):
        """Push a new peer into the heap"""

        self.order += 1
        peer.order = self.order + random.randint(0, self.size())
        heap.push(self, peer)

    def add_and_shuffle(self, peer):
        """Push a new peer into the heap and shuffle the heap"""
        self.push_peer(peer)

        r = random.randint(0, self.size() - 1)
        self.swap_order(peer.index, r)

    def peek_peer(self):
        """Return the top peer of the heap
        :return
            return the top peer if heap is not empty. Otherwise return None.
        """

        if not self.peers:
            return None

        return self.peers[0]

    def remove_peer(self, peer):
        """Remove the peer from the heap.

        Return: removed peer if peer exists. If peer's index is out of range,
        raise IndexError.
        """
        if peer.index < 0 or peer.index >= self.size():
            raise IndexError('Peer index is out of range')

        assert peer is self.peers[peer.index], "peer is not in the heap"

        return heap.remove(self, peer.index)

    def smallest_peer(self, predicate):
        """Return the smallest peer in the heap that matches the given
        predicate.

        :param predicate:
            Function that accepts an item from the heap and returns
            true or false.
        :returns:
            The first peer for which ``predicate`` returned true.
        """
        try:
            return self.peek(
                smallest(self, predicate),
            )
        except NoMatchError:
            return None

    def swap_order(self, index1, index2):
        if index1 == index2:
            return

        p1 = self.peers[index1]
        p2 = self.peers[index2]

        (p1.order, p2.order) = (p2.order, p1.order)

        heap.fix(self, p1.index)
        heap.fix(self, p2.index)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from tchannel.event import EventHook
from tchannel.zipkin import annotation
from tchannel.zipkin.tracers import DebugTracer
from tchannel.zipkin.tracers import TChannelZipkinTracer


class ZipkinTraceHook(EventHook):
    """generate zipkin-style span for tracing"""

    DEFAULT_RATE = 0.01

    def __init__(self, tchannel=None, dst=None, sample_rate=None):
        """Log zipkin style trace.

        :param tchannel:
            The tchannel instance to send zipkin trace spans
        :param dst:
            The destination to output trace information
        :param sample_rate:
            The sample_rate determines the probability that the trace span
            been sampled.
            The rate of sampling is in the range [0, 1] with 0.01 precision.
            By default it takes 100% sampling.
        """

        if tchannel:
            # TChannelZipkinTracer generates Base64-encoded span
            # and uploads to zipkin server
            self.tracer = TChannelZipkinTracer(tchannel)
        else:
            # DebugTracer generates json style span info and writes
            # to dst. By default it writes to stdout
            self.tracer = DebugTracer(dst)

        if sample_rate is None:
            sample_rate = self.DEFAULT_RATE

        assert 0 <= sample_rate <= 1
        self.rate = sample_rate
        self._check_point = self.rate * (1 << 64)

    def _lucky(self, id):
        return id < self._check_point

    def before_send_request(self, request):
        if not request.tracing.traceflags:
            return

        if not request.tracing.parent_span_id and not self._lucky(
            request.tracing.trace_id
        ):
            # disable trace
            request.tracing.traceflags = False
            return

        ann = annotation.client_send()
        request.tracing.annotations.append(ann)

    def before_receive_request(self, request):
        if not request.tracing.traceflags:
            return

        request.tracing.annotations.append(annotation.server_recv())

        caller_name = request.headers.get('cn')
        if caller_name:
            request.tracing.annotations.append(
                annotation.string('cn', caller_name),
            )

    def after_send_response(self, response):
        if not response.tracing.traceflags:
            return

        # send out a pair of annotations{server_recv, server_send} to zipkin
        ann = annotation.server_send()
        response.tracing.annotations.append(ann)
        self.tracer.record([(response.tracing, response.tracing.annotations)])

    def after_receive_response(self, request, response):
        if not response.tracing.traceflags:
            return

        # send out a pair of annotations{client_recv, client_send} to zipkin
        ann = annotation.client_recv()
        response.tracing.annotations.append(ann)
        self.tracer.record([(response.tracing, response.tracing.annotations)])

    def after_receive_error(self, request, error):
        if not error.tracing.traceflags:
            return

        ann = annotation.client_recv()
        error.tracing.annotations.append(ann)
        self.tracer.record([(error.tracing, error.tracing.annotations)])

    def after_send_error(self, error):
        if not error.tracing.traceflags:
            return

        ann = annotation.server_send()
        error.tracing.annotations.append(ann)
        self.tracer.record([(error.tracing, error.tracing.annotations)])



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from . import schemes
from . import transport as t

__all__ = ['Request']


class Request(object):
    """A TChannel request.

    This is sent by callers and received by registered handlers.

    :ivar body:
        The payload of this request. The type of this attribute depends on the
        scheme being used (e.g., JSON, Thrift, etc.).

    :ivar headers:
        A dictionary of application headers. This should be a mapping of
        strings to strings.

    :ivar transport:
        Protocol-level transport headers. These are used for routing over
        Hyperbahn.

        The most useful piece of information here is probably
        ``request.transport.caller_name``, which is the identity of the
        application that created this request.

    :ivar service:
        Name of the service being called. Inside request handlers, this is
        usually the name of "this" service itself. However, for services that
        simply forward requests to other services, this is the name of the
        target service.

    :ivar timeout:
        Amount of time (in seconds) within which this request is expected to
        finish.
    """

    # TODO move over other props from tchannel.tornado.request

    __slots__ = (
        'body',
        'headers',
        'service',
        'transport',
        'endpoint',
        'timeout',
    )

    def __init__(
        self,
        body=None,
        headers=None,
        transport=None,
        endpoint=None,
        service=None,
        timeout=None,
    ):
        self.body = body
        self.headers = headers
        self.transport = transport
        self.endpoint = endpoint
        self.service = service
        self.timeout = timeout


class TransportHeaders(object):
    """Request Transport Headers"""

    # TODO implement __repr__
    # TODO retry_flags should be woke up past a string

    __slots__ = (
        'caller_name',
        'claim_at_start',
        'claim_at_finish',
        'failure_domain',
        'retry_flags',
        'routing_delegate',
        'scheme',
        'speculative_exe',
        'shard_key',
    )

    def __init__(self,
                 caller_name=None,
                 claim_at_start=None,
                 claim_at_finish=None,
                 failure_domain=None,
                 retry_flags=None,
                 scheme=None,
                 speculative_exe=None,
                 shard_key=None,
                 routing_delegate=None):

        if scheme is None:
            scheme = schemes.RAW

        self.caller_name = caller_name
        self.claim_at_start = claim_at_start
        self.claim_at_finish = claim_at_finish
        self.failure_domain = failure_domain
        self.retry_flags = retry_flags
        self.routing_delegate = routing_delegate
        self.scheme = scheme
        self.speculative_exe = speculative_exe
        self.shard_key = shard_key

    @classmethod
    def from_dict(cls, data):
        return cls(
            caller_name=data.get(t.CALLER_NAME),
            claim_at_finish=data.get(t.CLAIM_AT_FINISH),
            claim_at_start=data.get(t.CLAIM_AT_START),
            failure_domain=data.get(t.FAILURE_DOMAIN),
            retry_flags=data.get(t.RETRY_FLAGS),
            routing_delegate=data.get(t.ROUTING_DELEGATE),
            scheme=data.get(t.SCHEME),
            shard_key=data.get(t.SHARD_KEY),
            speculative_exe=data.get(t.SPECULATIVE_EXE),
        )

    def to_dict(self):
        m = {}

        if self.caller_name is not None:
            m[t.CALLER_NAME] = self.caller_name

        if self.claim_at_start is not None:
            m[t.CLAIM_AT_START] = self.claim_at_start

        if self.claim_at_finish is not None:
            m[t.CLAIM_AT_FINISH] = self.claim_at_finish

        if self.failure_domain is not None:
            m[t.FAILURE_DOMAIN] = self.failure_domain

        if self.retry_flags is not None:
            m[t.RETRY_FLAGS] = self.retry_flags

        if self.routing_delegate is not None:
            m[t.ROUTING_DELEGATE] = self.routing_delegate

        if self.scheme is not None:
            m[t.SCHEME] = self.scheme

        if self.shard_key is not None:
            m[t.SHARD_KEY] = self.shard_key

        if self.speculative_exe is not None:
            m[t.SPECULATIVE_EXE] = self.speculative_exe

        return m



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import json
import logging
import random
import time

import tornado.gen
import tornado.ioloop

from ..errors import TimeoutError
from .response import StatusCode

EXPO_BASE = 1.4  # try this first
MAX_DELAY = 10  # sec
MAX_ATTEMPT = 7  # pow(1.4, 8) > 10
DELAY = 3 * 60 * 1000  # ms delay time for successful advertise
FIRST_ADVERTISE_TIME = 30  # sec

log = logging.getLogger('tchannel')


def _prepare_next_ad(attempt_counter):
    delay_time = random.uniform(
        0, min(MAX_DELAY, EXPO_BASE ** attempt_counter)
    )
    attempt_counter = min(attempt_counter, MAX_ATTEMPT)
    return attempt_counter, delay_time


@tornado.gen.coroutine
def _advertise(tchannel, service):
    response = None
    try:
        response = yield tchannel.request(service='hyperbahn').send(
            arg1='ad',  # advertise
            arg2='',
            arg3=json.dumps({
                'services': [
                    {
                        'serviceName': service,
                        'cost': 0,
                    }
                ]
            }),
            headers={'as': 'json'},
            retry_limit=0,
            ttl=2.0,
        )
    except Exception as e:  # Big scope to keep it alive.
        log.info('Failed to register with Hyperbahn: %s', e, exc_info=True)
    else:
        if response.code != StatusCode.ok:
            log.info('Failed to register with Hyperbahn: %s', response)
        else:
            log.info('Successfully registered with Hyperbahn')

    raise tornado.gen.Return(response)


@tornado.gen.coroutine
def _advertise_with_backoff(tchannel, service, timeout=None):
    # first advertise rule apply here.
    attempt_counter = 0
    start = time.time()

    while True:
        if timeout and time.time() - start > timeout:
            raise TimeoutError(
                "Failed to register with Hyperbahn due to timeout.",
            )

        response = yield _advertise(tchannel, service)

        if response is not None and response.code is StatusCode.ok:
            break

        attempt_counter += 1
        attempt_counter, delay_time = _prepare_next_ad(attempt_counter)

        yield tornado.gen.sleep(delay_time)

    raise tornado.gen.Return(response)


@tornado.gen.coroutine
def advertise(tchannel, service, routers=None, timeout=None, router_file=None):
    """Advertise with Hyperbahn.

    See :py:class:`tchannel.TChannel.advertise`.
    """
    timeout = timeout or FIRST_ADVERTISE_TIME
    if routers is not None and router_file is not None:
        raise ValueError(
            'Only one of routers and router_file can be provided.')

    if routers is None and router_file is not None:
        # should just let the exceptions fly
        try:
            with open(router_file, 'r') as json_data:
                routers = json.load(json_data)
        except (IOError, OSError, ValueError):
            log.exception('Failed to read seed routers list.')
            raise

    for router in routers:
        # We use .get here instead of .add because we don't want to fail if a
        # TChannel already knows about some of the routers.
        tchannel.peers.get(router)

    result = yield _advertise_with_backoff(
        tchannel, service, timeout=timeout
    )

    advertise_loop = tornado.ioloop.PeriodicCallback(
        lambda: _advertise_with_backoff(tchannel, service, timeout=None),
        DELAY,
    )
    advertise_loop.start()

    raise tornado.gen.Return(result)

advertize = advertise  # just in case



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import json

from tchannel.schemes import JSON


class JsonSerializer(object):
    name = JSON

    def serialize_header(self, headers):
        headers = headers or {}

        for k, v in headers.iteritems():
            if not (isinstance(k, basestring) and isinstance(v, basestring)):
                raise ValueError(
                    'headers must be a map[string]string (a shallow dict '
                    'where keys and values are strings)'
                )

        return json.dumps(headers)

    def deserialize_header(self, headers):
        if not headers:
            return {}
        return json.loads(headers)

    def deserialize_body(self, obj):
        return json.loads(obj)

    def serialize_body(self, obj):
        return json.dumps(obj)



from setuptools import find_packages, setup

import re

version = None
with open('tchannel/__init__.py', 'r') as f:
    for line in f:
        m = re.match(r'^__version__\s*=\s*(["\'])([^"\']+)\1', line)
        if m:
            version = m.group(2)
            break

if not version:
    raise Exception(
        'Could not determine version number from tchannel/__init__.py'
    )


setup(
    name='tchannel',
    version=version,
    author=', '.join([
        'Abhinav Gupta',
        'Aiden Scandella',
        'Bryce Lampe',
        'Grayson Koonce',
        'Junchao Wu',
    ]),
    author_email='abg@uber.com',
    description='Network multiplexing and framing protocol for RPC',
    license='MIT',
    url='https://github.com/uber/tchannel-python',
    keywords=['rpc'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(exclude=['tests', 'tests.*']),
    package_data={
        '': ['*.thrift'],
    },
    install_requires=[
        # stdlib backports, no constraints needed
        'contextlib2',
        'futures',

        # external deps
        'crcmod>=1,<2',
        'tornado>=4.2,<5',

        # tchannel deps
        'thriftrw>=0.4,<2',
        'threadloop>=1,<2',
    ],
    extras_require={
        'vcr': ['PyYAML', 'mock', 'wrapt'],
    },
    entry_points={
        'console_scripts': [
            'tcurl.py = tchannel.tcurl:start_ioloop'
        ]
    },
)
